import nanosite.util as util
import nanosite.build as build
import nanosite.server as server
import nanosite.packages as packages

import os
import json
import argparse

def is_in_nanosite_dir(path="."):
    if os.path.isfile(os.path.join(path, ".nanosite")):
        return True
    else:
        up = os.path.join("..", path)
        if util.same_path(up, path):  # reached root
            return False
        else:
            return is_in_nanosite_dir(up)

def get_cmdline_args():
    parser = argparse.ArgumentParser(prog="nanosite")
    parser.add_argument("action", nargs="?", default="",
                        help="options: build, serve, publish, " + \
                             "import, clean, delete")
    parser.add_argument("parameter", nargs="?", default="")
    parser.add_argument("--port", "-p", action="store", dest="port",
                        default="8000", type=int, help="set server port")
    parser.add_argument("-s", action="store", dest="site_dir",
                        default=os.getcwd(), help="specify site directory")
    parser.add_argument("-o", action="store", dest="output_dir",
                        default="output/", help="set output directory")
    parser.add_argument("-m", action="store", dest="meta_dir",
                        default="meta/", help="set meta directory")
    parser.add_argument("--set-package-url", action="store",
                        dest="package_url",
                        default=None, help="set package repository URL")
    parser.add_argument("--force", "-f", action="store_true", dest="force",
                        help="force actions")
    return parser.parse_args()

def setup_blank_site(top, ctx, meta):
    with open(".nanosite", "w") as f:
        json.dump({}, f)
    open("index.html+", "w").close()
    meta_dir = os.path.join(top, ctx["MetaDir"])
    os.makedirs(meta_dir, exist_ok=True)
    with open(os.path.join(meta_dir, "master.tmpl"), "w") as f:
        f.write("{{{content}}}")
    with open(os.path.join(meta_dir, "macros.py"), "w") as f:
        f.write('# macro("example", lambda ctx: fetch(ctx, "site.title"))\n')
    with open(os.path.join(meta_dir, "meta.json"), "w") as f:
        json.dump(meta, f, sort_keys=True, indent=2)
    if os.name == "nt":
        with open(os.path.join(meta_dir, "publish.bat"), "w") as f:
            f.write("rem This script is run during `nanosite publish`\n")
    else:
        publish_script_path = os.path.join(meta_dir, "publish")
        with open(publish_script_path, "w") as f:
            f.write("#!/bin/bash\n")
            f.write("# This script is run during `nanosite publish`\n")
        os.chmod(publish_script_path, 0o744)  # make executable

def setup_site_interactive(top, ctx):
    if util.prompt_YN("Would you like to set up a site in this directory?"):
        print("Enter a title for your site: ", end="")
        title = input()
        print("Enter a tagline for your site: ", end="")
        tagline = input()
        print("Enter the author name for your site: ", end="")
        author = input()
        meta = {"site": {"title": title, "tagline": tagline,
                         "author": author, "url": ""}}
        setup_blank_site(top, ctx, meta)
        print("Success! Generated site.")
    else:
        print("Canceled.")

def main():
    args = get_cmdline_args()
    action = args.action.lower()
    param = args.parameter

    site_exists = is_in_nanosite_dir(args.site_dir)

    ctx = {"OutputDir": args.output_dir, "MetaDir": args.meta_dir}
    if action == "":
        if args.package_url:  # trying to set package url
            packages.set_package_url(args.package_url, args.site_dir)
            print("Updated package URL.")
        else:
            if site_exists:  # default action: run server
                server.run_server(args.port, args.site_dir, ctx)
            else:
                setup_site_interactive(args.site_dir, ctx)
    elif not site_exists:
        print("No site in this directory.")
    elif action == "build" or action == "b":
        build.make_site(args.site_dir, ctx)
        print("Built site.")
    elif action == "clean" or action == "c":
        util.clean_output_dir(args.site_dir, args.output_dir)
    elif action == "delete" or action == "d":
        if util.prompt_YN("Are you sure you want to delete the site " +
                          "in this directory?"):
            util.delete_site_dir(args.site_dir)
            print("Deleted site.")
        else:
            print("Canceled.")
    elif action == "publish" or action == "p":
        util.publish_site(args.site_dir, args.meta_dir)
    elif action == "serve" or action == "s":
        server.run_server(args.port, args.site_dir, ctx)
    elif action == "import" or action == "i":
        if param:
            print("Installing package", param.lower())
            success, msg = packages.import_package(param, args.site_dir, ctx,
                                                   force=args.force)
            if success:
                print("Successfully installed package.")
            else:
                print(msg)
                print("Import failed.")
        else:
            print("Usage: nanosite import [package-name]")
    else:
        print("Unrecognized action. " +
              "Valid actions: build, serve, publish, import, clean, delete.")

if __name__ == "__main__":
    main()
    



import os
import markdown
import shutil
import subprocess

def same_path(a, b):
    return os.path.abspath(a).lower() == os.path.abspath(b).lower()

# compile markdown to HTML, returning (html, meta_info) tuple
def compile_markdown(md_text):
    md = markdown.Markdown(extensions=["markdown.extensions.meta"])
    html = md.convert(md_text)
    try:
        meta = {k: "".join(v) for k, v in md.Meta.items()} \
           if md.Meta is not None else {}
    except AttributeError:
        meta = {}
    return (html, meta)

def clean_output_dir(site_dir, output_dir):
    path = os.path.join(site_dir, output_dir)
    if os.path.isdir(path):
        shutil.rmtree(path)
        print("Cleaned output directory.")
    else:
        print("Nothing to clean.")
        
def delete_site_dir(top):
    for f in os.listdir(top):
        path = os.path.join(top, f)
        if os.path.isfile(path):
            os.unlink(path)
        elif os.path.isdir(path):
            shutil.rmtree(path)
    
def publish_site(top, meta_dir):
    cur_dir = os.getcwd()
    os.chdir(top)
    # run publish script from site top directory
    subprocess.call(os.path.join(top, meta_dir, "publish"))
    os.chdir(cur_dir)
    print("Finished running publish script.")

def prompt_YN(prompt):
    full_prompt = prompt + " [y/n] "
    print(full_prompt, end="")
    x = input()
    while x[0].lower() != "y" and x[0].lower() != "n":
        print("Invalid option. Type 'y' for yes and 'n' for no.")
        print(full_prompt, end="")
        x = input()
    return x[0].lower() == "y"



import nanosite.templates as templates
import nanosite.util as util

import os
from shutil import copyfile
from time import localtime
import json

def build_file(top, node, ctx):
    assert(node["is_file"])
    
    path = os.path.join(top, node["input_path"])
    root, ext = os.path.splitext(path)

    modified_files = []
    if ext.lower() in {".md", ".md+", ".html+"}:  # compile against templates
        # copy all node properties
        for k in node:
            ctx[k] = node[k]

        # fill local template
        template_path = node["template_path"]
        if template_path is not None:
            local_tmpl = templates.get_template(template_path)
            local_out_html = templates.fill_template(local_tmpl, ctx)
        else:
            local_out_html = node["content"]

        # get master template
        master_tmpl_path = os.path.join(top, ctx["MetaDir"], "master.tmpl")
        master_tmpl = templates.get_template(master_tmpl_path)

        # fill master template
        ctx["content"] = local_out_html
        out_html = templates.fill_template(master_tmpl, ctx)

        # remove escape symbols on delimiters
        out_html = templates.unescape_delimiters(out_html)

        relpath = os.path.relpath(root + ".html", top)
        out_path = os.path.join(top, ctx["OutputDir"], relpath)
        os.makedirs(os.path.dirname(out_path), exist_ok=True)
        # create output file in output folder
        with open(out_path, "w") as f:
            # get/create output path
            f.write(out_html)
        modified_files = [os.path.abspath(out_path)]
    elif ext.lower() == ".tmpl":  # no action
        pass
    elif ext.lower() == ".xml+":
        tmpl = templates.get_template(path)
        out_html = templates.fill_template(tmpl, ctx)
        
        # remove escape symbols on delimiters
        out_html = templates.unescape_delimiters(out_html)
        
        relpath = os.path.relpath(root + ".xml", top)
        out_path = os.path.join(top, ctx["OutputDir"], relpath)
        os.makedirs(os.path.dirname(out_path), exist_ok=True)
        
        with open(out_path, "w") as f:
            f.write(out_html)
        modified_files = [os.path.abspath(out_path)]
    else:   # copy file (make hard link on unix)
        relpath = os.path.relpath(path, top)
        out_path = os.path.join(top, ctx["OutputDir"], relpath)
        os.makedirs(os.path.dirname(out_path), exist_ok=True)
        # copy if src and dest are different (i.e. OutputDir != ".")
        if not util.same_path(path, out_path):
            if os.name == "nt":
                copyfile(path, out_path)
            else:
                # only copy file if it's new or newly modified 
                if not os.path.lexists(out_path) or \
                   os.path.getmtime(path) > os.path.getmtime(out_path):
                    #print("copying", path, "->", out_path)
                    copyfile(path, out_path)
            modified_files = [os.path.abspath(out_path)]
    return modified_files
    
def add_dirtree_file(top, path, ctx, template_path):
    root, ext = os.path.splitext(path)

    out_dict = {}
    new_ext = ext
    if ext.lower() == ".md" or ext.lower() == ".md+":
        html, meta = util.compile_markdown(open(path, "r").read())
        if ext.lower() == ".md":  # don't template plain .md files
            html = templates.escape_delimiters(html)
        out_dict = {"content": html}
        for k in meta:
            out_dict[k] = meta[k]
        new_ext = ".html"
        
        #ctx["content"] = html
        #ctx["meta"] = meta
    elif ext.lower() == ".html+" or ext.lower() == ".xml+":
        contents = open(path, "r").read()
        #local_out_html = fill_template(contents, ctx)
        out_dict = {"content": contents}#local_out_html}
 
        new_ext = new_ext[:-1]#".html"
        #ctx["content"] = local_out_html
    elif ext.lower() == ".tmpl":
        out_dict = {}
    else:  # copy file
        #relpath = os.path.relpath(path, top)
        #out_path = os.path.join(top, ctx["OutputDir"], relpath)
        #os.makedirs(os.path.dirname(out_path), exist_ok=True)
        #shutil.copyfile(path, out_path)
        out_dict = {}

    out_dict["is_file"] = True
    out_dict["input_path"] = os.path.relpath(path, top)
    out_dict["path"] = os.path.splitext(os.path.relpath(path, top))[0] + new_ext
    out_dict["template_path"] = template_path
    out_dict["date"] = localtime(os.path.getmtime(path))
    return out_dict

def make_dirtree(top, path, ctx, template_path=None):
    files = []
    dirs = []
    tree = {}
    for subdir in os.listdir(path):
        if subdir[0] != ".":  # don't include hidden files
            subpath = os.path.join(path, subdir)
            if os.path.isfile(subpath):
                # update template path if there is a template in this folder
                root, ext = os.path.splitext(os.path.basename(subpath))
                if root.lower() == "template" and ext.lower() == ".tmpl":
                    template_path = subpath
                else:
                    files.append((subdir, subpath))
            elif os.path.isdir(subpath):
                rp = os.path.abspath(subpath)
                md = os.path.join(top, ctx["MetaDir"])
                od = os.path.join(top, ctx["OutputDir"])
                if not (util.same_path(rp, md) or util.same_path(rp, od)):
                    dirs.append((subdir, subpath))

    for short_path, full_path in dirs:
        path_name = short_path.lower()
        tree[path_name] = make_dirtree(top, full_path, ctx, template_path)
        
    for short_path, full_path in files:
        name = short_path.replace(".", "_")
        if name in tree: name = name.replace("_", "__")
        tree[name] = add_dirtree_file(top, full_path, dict(ctx), template_path)

    return tree

# traverse tree to build all site files
def compile_dirtree(top, tree, ctx):
    modified_files = []
    for k, node in tree.items():
        if "is_file" in node and node["is_file"]:
            mf = build_file(top, node, dict(ctx))
        else:
            mf = compile_dirtree(top, node, ctx)
        modified_files += mf
    return modified_files
            
def load_meta(top, ctx):
    meta_path = os.path.join(top, ctx["MetaDir"], "meta.json")
    if os.path.isfile(meta_path):
        with open(meta_path, "r") as f:
            meta = json.loads(f.read())
        for key in meta:
            ctx[key] = meta[key]
    return ctx
    
def register_macros(top, ctx):
    # functions to be exposed to macros.py
    def macro(s, fun):
        ctx[s] = fun
    def fetch(ctx, key):
        return util.ctx_fetch(ctx, key)
    pgm_path = os.path.join(top, ctx["MetaDir"], "macros.py")
    if os.path.isfile(pgm_path):
        with open(pgm_path, "r") as f:
            pgm = f.read()
        exec(pgm, dict(list(globals().items()) + list(locals().items())))
    return ctx

# returns array of absolute paths of the files modified
def make_site(top, ctx):
    templates.clear_template_cache()

    # load meta variables
    ctx = load_meta(top, ctx)

    # register macros
    ctx = register_macros(top, ctx)

    # create dirtree
    tree = make_dirtree(top, top, ctx)
    
    # copy dirtree to context
    for k in tree:
        ctx[k] = tree[k]
        
    # fill master template in order to register global macros
    master_tmpl_path = os.path.join(top, ctx["MetaDir"], "master.tmpl")
    templates.fill_template(templates.get_template(master_tmpl_path), ctx)

    # compile dirtree
    mf = compile_dirtree(top, tree, ctx)
    return mf



import nanosite.util as util
import nanosite.build as build

import os
from datetime import datetime
from traceback import print_exc as trace_print_exc
from traceback import format_exc as trace_format_exc
from time import sleep
from http.server import SimpleHTTPRequestHandler
from socketserver import TCPServer
from threading import Thread
from atexit import register as atexit_register

def update(top, ctx, handler):
    def last_update_time(walk, ignore=[], last_t=None):
        # getmtime may fail if files changed since os.walk call
        def mtime_or_zero(path):
            try:
                return os.path.getmtime(path)
            except FileNotFoundError:
                return 0
            
        cur_max = 0
        for path, dirs, files in walk:
            fs = [(os.path.join(path, f),
                   mtime_or_zero(os.path.join(path, f))) for f in files \
                  if list(filter(lambda x:
                                 util.same_path(x, os.path.join(path, f)),
                                 ignore)) == [] \
                  and f[0] != "."]
            for n, t in fs:
                if last_t is not None and t > last_t: print(n, "modified")
            if fs != []:
                fs = list(zip(*fs))[1]
                cur_max = max(cur_max, max(fs))
        return cur_max
    
    needs_update = True
    last_t = None
    last_walk = None
    while True:
        if needs_update:
            needs_update = False
            try:
                mf = build.make_site(top, ctx)
                handler.error = None
                print("[" + str(datetime.now()) + "]", "Built site.")
            except:
                last_t = None
                mf = []
                trace_print_exc()
                handler.error = trace_format_exc()
        walk = sorted(list(os.walk(top)))
        t = last_update_time(walk, mf, last_t)
        sleep(0.5)
        # file updated or dirtree changed
        if last_t is not None and (t != last_t or walk != last_walk):
            needs_update = True
        last_t = t
        last_walk = walk

def run_server(port, site_dir, ctx):
    output_dir = ctx["OutputDir"]

    # start server
    class RequestHandler(SimpleHTTPRequestHandler):
        def do_GET(self):
            if hasattr(self, "error") and self.error is not None:
                self.send_response(200, 'OK')
                self.send_header('Content-type', 'html')
                self.end_headers()
                self.wfile.write(bytes(self.error, 'UTF-8'))
            else:
                super().do_GET()
        def translate_path(self, path):
            return os.path.join(site_dir, output_dir, path[1:])
    handler = RequestHandler
    httpd = TCPServer(("", port), handler)
    atexit_register(lambda: httpd.shutdown())
    
    # start update thread
    thread = Thread(target=update, args=(site_dir, ctx, handler))
    thread.daemon = True
    thread.start()
    
    print("Serving at http://localhost:" + str(port) + "/")
    httpd.serve_forever()



from templates import fill_template

# fill template
assert fill_template("hello {{key}}!", {"key": "world"}) == "hello world!"
assert(fill_template("{{a.b.c}}", {"a": {"b": {"c": 3}}}) == "3")
assert(fill_template("{{#macro test a}}{{#if a}}A{{#else}}B{{#endif}}{{#endmacro}}{{test 1}}{{test 0}}", {}) == "AB")
assert(fill_template("{{#macro echo a}}{{a}}{{#endmacro}}{{#define x echo \"hello\"}}{{x}}", {}) == "hello")
assert(fill_template("{{#define x \"hello world\"}}{{x}}", {}) == "hello world")
assert(fill_template("{{#for x in items}}{{x}}.{{#endfor}}", {"items": [1, 2, 3]}) == "1.2.3.")
assert(fill_template("{{#define x \"hello world\"}}{{x}}", {}) == "hello world")
assert(fill_template("{{#if no}}nope{{#else}}{{#for x in items}}{{x}}.{{#endfor}}{{#endif}}", {"no": False, "items": [1, 2, 3]}) == "1.2.3.")
assert(fill_template("{{#if a}}A{{#if b}}B{{#elif c}}C{{#else}}(!B!C){{#endif}}(Y){{#else}}(!A){{#endif}}", {"a": True, "b": True, "c": True}) == "AB(Y)")
assert(fill_template("{{#if a}}A{{#if b}}B{{#elif c}}C{{#else}}(!B!C){{#endif}}(Y){{#else}}(!A){{#endif}}", {"a": True, "b": True, "c": False}) == "AB(Y)")
assert(fill_template("{{#if a}}A{{#if b}}B{{#elif c}}C{{#else}}(!B!C){{#endif}}(Y){{#else}}(!A){{#endif}}", {"a": True, "b": False, "c": True}) == "AC(Y)")
assert(fill_template("{{#if a}}A{{#if b}}B{{#elif c}}C{{#else}}(!B!C){{#endif}}(Y){{#else}}(!A){{#endif}}", {"a": True, "b": False, "c": False}) == "A(!B!C)(Y)")
assert(fill_template("{{#if a}}A{{#if b}}B{{#elif c}}C{{#else}}(!B!C){{#endif}}(Y){{#else}}(!A){{#endif}}", {"a": False, "b": True, "c": True}) == "(!A)")
assert(fill_template("{{#if a}}A{{#if b}}B{{#elif c}}C{{#else}}(!B!C){{#endif}}(Y){{#else}}(!A){{#endif}}", {"a": False, "b": True, "c": False}) == "(!A)")
assert(fill_template("{{#if a}}A{{#if b}}B{{#elif c}}C{{#else}}(!B!C){{#endif}}(Y){{#else}}(!A){{#endif}}", {"a": False, "b": False, "c": True}) == "(!A)")
assert(fill_template("{{#if a}}A{{#if b}}B{{#elif c}}C{{#else}}(!B!C){{#endif}}(Y){{#else}}(!A){{#endif}}", {"a": False, "b": False, "c": False}) == "(!A)")



import nanosite.templates as templates
import nanosite.util as util

import os
from zipfile import ZipFile
import json
from urllib.request import urlopen
from urllib.parse import urljoin

DefaultPackageURL = "http://wanganzhou.com/nanosite/packages/"

def set_package_url(url, top):
    if url and url[-1] != "/": url += "/"  # add trailing slash if necessary
    dot_nanosite = None
    with open(os.path.join(top, ".nanosite"), "r") as f:
        dot_nanosite = json.loads(f.read())
        dot_nanosite["package-url"] = url
    if dot_nanosite:
        with open(os.path.join(top, ".nanosite"), "w") as dnf:
            json.dump(dot_nanosite, dnf, indent=2)

# name: package name
# package_url: package repo URL
# returns whether download succeeded
def download_package(name, package_url, dest):
    filename = name + ".zip"
    url = urljoin(package_url, filename)
    print("Downloading package", name, "from", url)
    try:
        with urlopen(url) as response, \
         open(filename, "wb") as out_file:
            out_file.write(response.read())
        return True
    except:
        return False

# f: ZipFile object of package file
# dot_nanosite: contents of .nanosite file (loaded from JSON)
# force: whether to force override
def install_package(name, f, dot_nanosite, top, ctx, force=False):
    with f.open("rules.json", "rU") as rules_file:
        rules = json.loads(rules_file.read().decode("utf-8"))
        files = rules["files"] if "files" in rules else {}
        dependencies = rules["dependencies"] if "dependencies" in rules else []

        # check dependencies
        installed_packages = dot_nanosite["installed-packages"]
        for dependency in dependencies:
            if dependency not in installed_packages:
                print("Installing dependency", dependency)
                success, msg = import_package(dependency, top, ctx)
                if not success:
                    return False, "Unable to install dependency " + dependency

        # preliminarily go through files, prompt user to overwrite if needed
        overwritten_files = []
        for filename in files:
            rule = files[filename]
            dest = os.path.join(top, templates.fill_template(rule["dest"],
                                                             ctx))
            # an overwrite will occur
            if os.path.isfile(dest) and "w" in rule["action"].lower():
                overwritten_files.append(dest)
        if overwritten_files:
            print("The following files will be overwritten:")
            for fn in overwritten_files:
                print(" -", fn)
            if not util.prompt_YN("Are you sure you want to continue?"):
                return False, "Canceled."
            
        # go through files
        for filename in files:
            rule = files[filename]
            dest = os.path.join(top, templates.fill_template(rule["dest"],
                                                             ctx))
            
            os.makedirs(os.path.dirname(dest), exist_ok=True)
            with f.open(filename, "r") as src_file:
                with open(dest, rule["action"]) as dest_file:
                    dest_file.write(src_file.read().decode("utf-8"))

        # add to list of installed packages
        if name not in installed_packages:
            installed_packages.append(name)
        dot_nanosite["installed-packages"] = installed_packages
        with open(os.path.join(top, ".nanosite"), "w") as dnf:
            json.dump(dot_nanosite, dnf, indent=2)
    return True, ""
    
# returns (success (bool), status (string))
# force: whether to force override
def import_package(name, top, ctx, force=False):
    name = name.lower()
    
    # setup package info in .nanosite file if first time
    with open(os.path.join(top, ".nanosite"), "r") as f:
        dot_nanosite = json.loads(f.read())
        if "installed-packages" not in dot_nanosite:
            dot_nanosite["installed-packages"] = []
        if "package-url" not in dot_nanosite:
            dot_nanosite["package-url"] = DefaultPackageURL
        installed_packages = dot_nanosite["installed-packages"]
        package_url = dot_nanosite["package-url"]

        # remove trailing slashes in MetaDir and OutputDir if needed
        if "MetaDir" in ctx and ctx["MetaDir"][-1] == "/":
            ctx["MetaDir"] = ctx["MetaDir"][:-1]
        if "OutputDir" in ctx and ctx["OutputDir"][-1] == "/":
            ctx["OutputDir"] = ctx["OutputDir"][:-1]

        # check if package already installed
        if name in installed_packages and not force:
            return False, name + " already installed"

        # check if package file is in site top. If not, download it
        filename = name + ".zip"
        path = os.path.join(top, filename)
        downloaded_package = False
        if not os.path.isfile(path):
            downloaded_package = True
            success = download_package(name, package_url, path)
            if not success: return False, "Could not find package"

        # install package
        with ZipFile(path, "r") as f:
            success, msg = install_package(name, f, dot_nanosite, top, ctx,
                                           force=force)

        # delete package if downloaded
        if downloaded_package:
            os.remove(path)

        return success, msg



import re
from html import escape as escape_HTML

# escape templating delimiters
def escape_delimiters(s):
    return s.replace("{{", "\\{{").replace("}}", "\\}}")

def unescape_delimiters(s):
    return s.replace("\\{{", "{{").replace("\\}}", "}}")

# fetch key, possibly nested thru dot notation
def ctx_fetch(ctx, line):
    key, params = (line.split(" ", 1) + [""])[:2]
    parts = key.split(".", 1)
    is_attr = hasattr(ctx, parts[0])
    if parts[0] in ctx or is_attr:
        if len(parts) == 1:
            value = None
            try:  # try getting it as a dictionary entry first
                value = ctx[parts[0]]
            except:  # otherwise try getting it as an attribute
                value = getattr(ctx, parts[0])
            # if it's a macro, call it with parameter [ctx]
            if callable(value):
                value = value(ctx, *tokenize_params(ctx, params))
            return value
        else:
            return ctx_fetch(ctx[parts[0]], parts[1])
    else:
        return None

# get params from space-separated parameter list
# if greedy, then context grabs consume all of the remaining parameters
#   (this is for macro calls with parameters)
def tokenize_params(ctx, params, greedy=False):
    out = []
    in_string = False
    tokens = params.split(" ")
    for i, token in enumerate(tokens):
        if token != "":
            if token[0] == '"':  # begin string
                in_string = True
                token = token[1:]
                out.append("")
            if in_string:  # continue string
                if token[-1] == '"': 
                    token = token[:-1]
                    in_string = False
                out[-1] += token + (" " if in_string else "")
            else:
                try:  # int 
                    out.append(int(token))
                except ValueError:  # float
                    try:
                        out.append(float(token))
                    except ValueError:  # identifier (so get object from ctx)
                        if greedy: token = " ".join(tokens[i:])
                        out.append(ctx_fetch(ctx, token))
                        if greedy: break
    return out

template_cache = {}
def clear_template_cache():
    global template_cache
    template_cache = {}
    
# get template from path (cached)
def get_template(path):
    global template_cache
    if path in template_cache:
        return template_cache[path]
    else:
        tmpl = open(path, "r").read()
        template_cache[path] = tmpl
        return tmpl

# fill template according to rules in doc.txt
# ctx is modified according to new variable bindings
def fill_template(tmpl, ctx):
    def make_define(ctx, cmd):
        ctx[cmd[1]] = tokenize_params(ctx, " ".join(cmd[2:]), greedy=True)[0]
                
    # get part before delim and part after
    def get_chunk(s, delim):
        #a = s.split(delim, 1)
        a = re.split(delim, s, 1)
        return (a[0], "") if len(a) <= 1 else a
    rest = tmpl
    out = ""
    seeking = None
    seek_depth = 0
    depth_if = 0
    depth_for = 0
    block_accum = ""
    for_variable = ""
    for_collection = None
    macro_name = ""
    macro_param_names = []
    while len(rest) > 0:
        # match {{, not \{{, \{{{
        cur, rest = get_chunk(rest, "(?<!\\\\)(?<!{){{")
        #print("}}", cur, "{{", rest, seeking, seek_depth, depth_if)
        #print()
        if seeking is None:
            out += cur
        else:
            block_accum += cur
        key, rest = get_chunk(rest, "(?<!\\\\)(?<!})}}")
        key = key.strip()
        cmd = key.split()
        if cmd: cmd[0] = cmd[0].lower()

        #print("{{", key, "}}", rest, seeking, seek_depth, depth_if)
        #print()
        #print()

        if seeking is not None:
            run_for_block = False
            create_macro = False
            if cmd[0] == "#if":
                depth_if += 1
            elif cmd[0] == "#else":
                if "#else" in seeking and seek_depth == depth_if:
                    #print("FOUND MATCHING ELSE", seek_depth, depth_if)
                    seeking = None
            elif cmd[0] == "#elif":
                if "#elif" in seeking and seek_depth == depth_if:
                    if ctx_fetch(ctx, cmd[1]):
                        #print("FOUND MATCHING ELIF", seek_depth, depth_if)
                        seeking = None
            elif cmd[0] == "#endif":
                if "#endif" in seeking and seek_depth == depth_if:
                    #print("FOUND MATCHING ENDIF", seek_depth, depth_if)
                    seeking = None
                depth_if -= 1
            elif cmd[0] == "#for":
                depth_for += 1
            elif cmd[0] == "#endfor":
                if "#endfor" in seeking and seek_depth == depth_for:
                    seeking = None
                    run_for_block = True
                depth_for -= 1
            elif cmd[0] == "#macro":
                raise Exception("#macro declared inside of another block")
            elif cmd[0] == "#endmacro":
                seeking = None
                create_macro = True
            elif cmd[0] == "#define":
                make_define(ctx, cmd)
            if run_for_block:
                orig = None  # save original context binding
                if for_variable in ctx:
                    orig = ctx["for_variable"]

                for item in for_collection:
                    ctx[for_variable] = item
                    out += fill_template(block_accum, ctx)
                if orig is None:  # restore original context binding
                    if for_variable in ctx:
                        ctx.pop(for_variable)
                else:
                    ctx[for_variable] = orig
            elif create_macro:
                def macro_func(ctx, *params):
                    origs = {}  # saves original bindings
                    for name, value in zip(macro_param_names, params):
                        if name in ctx: origs[name] = ctx[name]
                        ctx[name] = value
                    out = fill_template(block_accum, ctx)
                    for key in origs:  # restore bindings
                        ctx[key] = origs[key]
                    return out
                ctx[macro_name] = macro_func
            else:
                block_accum += "{{" + key + "}}"
        else:
            if cmd == [] or cmd[0] == "":
                pass
            elif cmd[0] == "#if":
                depth_if += 1
                if not ctx_fetch(ctx, cmd[1]):  # if ctx[key] evaluates to True
                    seeking = {"#elif", "#else", "#endif"}
                    seek_depth = depth_if
            elif cmd[0] == "#else" or cmd[0] == "#elif":
                if depth_if == 0:
                    raise Exception(cmd[0] + " without #if")
                seeking = {"#endif"}
                seek_depth = depth_if
            elif cmd[0] == "#endif":
                depth_if -= 1
            elif cmd[0] == "#for":
                depth_for += 1
                block_accum = ""
                seeking = {"#endfor"}
                seek_depth = depth_for
                for_variable = cmd[1]
                for_collection = ctx_fetch(ctx, " ".join(cmd[3:]))
            elif cmd[0] == "#endfor":
                depth_for -= 1
            elif cmd[0] == "#macro":
                block_accum = ""
                seeking = {"#endmacro"}
                macro_name = cmd[1]
                macro_param_names = cmd[2:]
            elif cmd[0] == "#define":
                make_define(ctx, cmd)
            else:
                if key[0] == "{":
                    val = str(ctx_fetch(ctx, key[1:].strip()))
                    _, rest = get_chunk(rest, "}") # get matching close brace
                else:
                   val = escape_HTML(str(ctx_fetch(ctx, key.strip())))
                out += fill_template(val, ctx)
                
    if depth_if > 0:
        raise Exception("#if without #endif")
    elif depth_if < 0:
        raise Exception("#endif without #if")
    if depth_for > 0:
        raise Exception("#for without #endfor")
    elif depth_for < 0:
        raise Exception("#endfor without #for")
    return out



from setuptools import setup

setup(
    name = "nanosite",
    author = "Andrew Wang",
    author_email = "azw7@cornell.edu",
    url = "https://github.com/qema/nanosite",
    version = "0.1.3",
    packages = ["nanosite"],
    install_requires = ["markdown"],
    entry_points = {
        "console_scripts": [
            "nanosite = nanosite:main"
            ]
        }
)
